import numpy as np
cimport numpy as np

from roc.rpl.packet_structure.data import Data
from roc.rpl.packet_structure.data cimport Data


cpdef tuple rpw_status(Data data, unsigned int start):
    """
    Return tuple of values from the RPW status command.
    """
    cdef:
        unsigned char bias_on_off
        unsigned char lfr_on_off
        unsigned char tds_on_off
        unsigned char thr_on_off
        unsigned char ant1_on_off
        unsigned char ant2_on_off
        unsigned char ant3_on_off
        unsigned char scm_on_off
        unsigned char bias3
        unsigned char bias2
        unsigned char bias1
        unsigned char hv
        unsigned char m_lfr
        unsigned char c_lfr
        unsigned char m_tds
        unsigned int tmp

    tmp = start + 1

    bias_on_off = data.u8(tmp, 7, 1)
    lfr_on_off = data.u8(tmp, 6, 1)
    tds_on_off = data.u8(tmp, 5, 1)
    thr_on_off = data.u8(tmp, 4, 1)
    ant1_on_off = data.u8(tmp, 3, 1)
    ant2_on_off = data.u8(tmp, 2, 1)
    ant3_on_off = data.u8(tmp, 1, 1)
    scm_on_off = data.u8(tmp, 0, 1)

    bias3 = data.u8(start, 7, 1)
    bias2 = data.u8(start, 6, 1)
    bias1 = data.u8(start, 5, 1)
    hv = data.u8(start, 4, 1)
    m_lfr = data.u8(start, 3, 1)
    c_lfr = data.u8(start, 2, 1)
    m_tds = data.u8(start, 1, 1)

    return (
        m_tds, c_lfr, m_lfr, hv, bias1, bias2, bias3, scm_on_off,
        ant3_on_off, ant2_on_off, ant1_on_off, thr_on_off, tds_on_off,
        lfr_on_off, bias_on_off
    )


cpdef tuple do_analysis(Data data, unsigned int start):
    """
    Return tuple of values from the DoAnalysis command.
    """
    cdef:
        unsigned char ch1
        unsigned char ch2
        unsigned char eos
        unsigned char mod
        unsigned char it

    # read parameters
    ch1 = data.u8(start, 7, 1)
    ch2 = data.u8(start, 6, 1)
    eos = data.u8(start, 5, 1)
    mod = data.u8(start, 4, 1)
    it = data.u8(start, 0, 4)

    return it, mod, eos, ch2, ch1


cpdef tuple input_setup(Data data, unsigned int start):
    """
    Return a tuple of values from the input setup command.
    """
    cdef:
        unsigned char sensor_tnr_1
        unsigned char sensor_tnr_2
        unsigned char fe
        unsigned int tmp

    tmp = start + 1

    # read parameters
    # N.B. Bit order of tnr1 and tnr2 sensors
    # are inverted w.r.t. the doc.
    # (RPW-SYS-MEB-THR-TN-000914-LES_Iss03_Rev04.pdf)
    sensor_tnr_2 = data.u8(tmp, 0, 4)
    sensor_tnr_1 = data.u8(tmp, 4, 4)
    fe = data.u16(start, 6, 2)

    return fe, sensor_tnr_2, sensor_tnr_1


cpdef tuple temperatures_command(Data data, unsigned int start):
    """
    Read temperatures command block from micro command.
    """
    cdef:
        unsigned char analog
        unsigned char ant1
        unsigned char ant2
        unsigned char ant3

    analog = data.u8(start, 0, 8)
    ant1 = data.u8(start + 1, 0, 8)
    ant2 = data.u8(start + 2, 0, 8)
    ant3 = data.u8(start + 3, 0, 8)

    return analog, ant1, ant2, ant3


cpdef tuple voltages(Data data, unsigned int start):
    """
    Read voltages from the command.
    """
    cdef:
        unsigned char minus5
        unsigned char plus5
        unsigned char plus12

    minus5 = data.u8(start, 0, 8)
    plus5 = data.u8(start + 1, 0, 8)
    plus12 = data.u8(start + 2, 0, 8)

    return minus5, plus5, plus12


cpdef list voltages_dtype():
    """
    Return voltages dtype.
    """
    return [
        ("minus5", "uint8"), ("plus5", "uint8"), ("plus12", "uint8")
    ]


cpdef list temperatures_dtype():
    """
    Return temperatures dtype.
    """
    return [
        ("analog", "uint8"),
        ("ant1", "uint8"),
        ("ant2", "uint8"),
        ("ant3", "uint8"),
    ]


cpdef list rpw_status_dtype():
    """
    Return the RPW status command dtype.
    """
    return [
        ("m_tds", "uint8"), ("c_lfr", "uint8"), ("m_lfr", "uint8"),
        ("hv", "uint8"), ("bias1", "uint8"), ("bias2", "uint8"),
        ("bias3", "uint8"), ("scm_on_off", "uint8"),
        ("ant3_on_off", "uint8"), ("ant2_on_off", "uint8"),
        ("ant1_on_off", "uint8"), ("thr_on_off", "uint8"),
        ("tds_on_off", "uint8"), ("lfr_on_off", "uint8"),
        ("bias_on_off", "uint8")
    ]


cpdef list input_setup_dtype():
    return [
        ("fe", "uint8"), ("sensor_tnr_2", "uint8"), ("sensor_tnr_1", "uint8")
    ]


cpdef list do_analysis_dtype():
    return [
        ("it", "uint8"), ("mod", "uint8"), ("eos", "uint8"),
        ("ch2", "uint8"), ("ch1", "uint8"),
    ]

cpdef unsigned int average_nr(unsigned int average_index):
    average_map = [16, 32, 64, 128]
    return average_map[average_index]





from libc.stdint cimport uint8_t
from libc.stdint cimport uint16_t
from libc.stdint cimport uint32_t
from libc.stdint cimport uint64_t

import numpy as np
cimport numpy as np

from poppy.core.logger import logger

from roc.rap.tasks.utils import h5_group_to_dict

from roc.rpl.time import Time
from roc.idb.converters.cdf_fill import fill_records

__all__ = ["MAMP", "set_mamp"]

# Numpy array dtype for TDS MAMP L1 survey data
# Name convention is lower case
MAMP = [
    ("epoch", "uint64"),
    ("scet", "float64"),
    ("acquisition_time", ("uint32", 2)),
    ("synchro_flag", "uint8"),
    ("quality_flag", "uint8"),
    ("quality_bitmask", "uint8"),
    ("survey_mode", "uint8"),
    ("bia_status_info", ("uint8", 6)),
    ("dec_rate", "uint8"),
    ("hf_data_artefacts", ("uint8", 5)),
    ("rpw_status_info", ("uint8", 8)),
    ("input_config", "uint32"),
    ("snapshot_seq_nr", "uint16"),
    ("channel_status_info", ("uint8", 4)),
    ("waveform_data", ("int16", 4)),
]

cpdef np.ndarray set_mamp(object l0, object task):
    """
    Extract TM_TDS_SCIENCE_NORMAL_MAMP data

    :param l0: h5py.h5 object containing the RPW TM/TC packets
    :param task: POPPy task
    :return: numpy array with TDS mamp data
    """

    return Mamp().set_mamp(l0, task)

cdef class Mamp:

    cdef public np.ndarray data
    cdef public Py_ssize_t index

    def __init__(self):
        # Initialize mamp data array
        self.data = np.empty(
            0,
            dtype=MAMP,
        )

        # initialize the position in the data array
        self.index = 0

    cpdef np.ndarray set_mamp(self, object l0, object task):

        cdef:
            Py_ssize_t packet
            Py_ssize_t size
            Py_ssize_t sample
            list void_list
            str name
            object tm

        # name of the packet
        name = "TM_TDS_SCIENCE_NORMAL_MAMP"
        #logger.debug("Get data for " + name)

        # get the TDS MAMP packet data
        if name in l0["TM"]:
            tm = l0["TM"][name]["source_data"]
        else:
            # rest of the code inside the with statement will not be called
            return self.data

        # Make sure to use dictionary instead of H5 group
        # (Increase read speed)
        tm = h5_group_to_dict(tm)

        # suppose for now that we have the good size for the data
        size = np.sum(tm["PA_TDS_MAMP_SAMP_PER_CH"])
        #logger.debug("Number of samples: {0}".format(size))

        # create the data array with good size
        self.data = np.empty(size, dtype=MAMP)
        fill_records(self.data)

        # get the number of packets
        size = tm["PA_TDS_ACQUISITION_TIME"].shape[0]
        #logger.debug("Number of packets: {0}".format(size))

        # survey mode to normal
        self.data["survey_mode"][:] = 0

        # loop over packets
        with nogil:
            for packet in range(size):
                with gil:
                    # loop over samples per channel
                    # defined in the variable PA_TDS_MAMP_SAMP_PER_CH
                    void_list = [
                        self._fill_mamp_data(tm, packet, sample)
                        for sample in range(tm["PA_TDS_MAMP_SAMP_PER_CH"][packet])
                    ]

        return self.data

    cdef _fill_mamp_data(self, object tm,
                         Py_ssize_t packet,
                         Py_ssize_t sample,
                         ):

        cdef:
            Py_ssize_t channel
            uint8_t counter_ch
            uint16_t snapshot_nr
        # populate the data array

        # Time variables
        self.data["acquisition_time"][self.index, :] = (
            tm["PA_TDS_ACQUISITION_TIME"][packet, 0],
            tm["PA_TDS_ACQUISITION_TIME"][packet, 1],
        )
        self.data["epoch"][self.index] = Time().obt_to_utc(
            np.array([self.data["acquisition_time"][self.index, :]]),
            to_tt2000=True)
        self.data["synchro_flag"][self.index] = tm[
            "PA_TDS_ACQUISITION_TIME"][packet, 2]
        self.data["scet"][self.index] = Time.cuc_to_scet(
            np.array([self.data["acquisition_time"][self.index, :]]))

        # dec rate
        self.data["dec_rate"][self.index] = tm[
            "PA_TDS_N_MAMP_DEC_RATE"][packet]

        # put the sampling rate
#        self.data["sampling_rate"][self.index] = tm[
#            "PA_TDS_MAMP_BASE_SAMP_RATE"][packet]

        # channel status info
        self.data["channel_status_info"][self.index] = (
            tm["PA_TDS_N_MAMP_ADC_CH1"][packet],
            tm["PA_TDS_N_MAMP_ADC_CH2"][packet],
            tm["PA_TDS_N_MAMP_ADC_CH3"][packet],
            tm["PA_TDS_N_MAMP_ADC_CH4"][packet],
        )

        # data artefacts
        self.data["hf_data_artefacts"][self.index] = (
            tm["PA_TDS_HF_ART_BUF_OVERFLOW"][packet],
            tm["PA_TDS_HF_ART_ADC1_OR"][packet],
            tm["PA_TDS_HF_ART_ADC2_OR"][packet],
            tm["PA_TDS_HF_ART_ADC3_OR"][packet],
            tm["PA_TDS_HF_ART_ADC4_OR"][packet],
        )

        # filter coefs
#        self.data["filter_coefs"][self.index] = tm[
#            "PA_TDS_FILTER_COEFS"][packet]

        # RPW status information
        self.data["rpw_status_info"][self.index] = (
            tm["PA_TDS_THR_OFF"][packet],
            tm["PA_TDS_LFR_OFF"][packet],
            tm["PA_TDS_ANT1_OFF"][packet],
            tm["PA_TDS_ANT2_OFF"][packet],
            tm["PA_TDS_ANT3_OFF"][packet],
            tm["PA_TDS_SCM_OFF"][packet],
            tm["PA_TDS_HF_ART_MAG_HEATER"][packet],
            tm["PA_TDS_HF_ART_SCM_CALIB"][packet],
        )

        # information for bias
        self.data["bia_status_info"][self.index] = (
            tm["PA_BIA_ON_OFF"][packet],
            tm["PA_BIA_MODE_BIAS3_ENABLED"][packet],
            tm["PA_BIA_MODE_BIAS2_ENABLED"][packet],
            tm["PA_BIA_MODE_BIAS1_ENABLED"][packet],
            tm["PA_BIA_MODE_HV_ENABLED"][packet],
            tm["PA_BIA_MODE_MUX_SET"][packet],
        )

        # snapshot number
        snapshot_nr = tm["PA_TDS_SNAPSHOT_NR"][packet]

        # snapshot number
        self.data["snapshot_seq_nr"][self.index] = snapshot_nr

        # input config
        self.data["input_config"][self.index] = tm[
            "PA_TDS_SWF_HF_CH_CONF"][packet]

        # # initialization of the counter of activated channel to be able
        # # to place read data correctly inside the block
        counter_ch = 0

        # # loop over channels and add data
        for channel in range(4):
            if self.data["channel_status_info"][self.index][channel] == 1:
                self.data["waveform_data"][self.index, channel] = tm[
                    "PA_TDS_MAMP_DATA_BLK"
                ][
                    packet,
                    sample + counter_ch *
                    tm["PA_TDS_MAMP_SAMP_PER_CH"][packet],
                ]
                counter_ch += 1

        # increment the self.index
        self.index += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pandas as pd

from poppy.core.generic.metaclasses import Singleton
from poppy.core.db.connector import Connector
from poppy.core.logger import logger

from roc.rap.constants import PIPELINE_DATABASE, TC_HFR_LIST_PAR

__all__ = ['HfrListMode']


class HfrListMode(object, metaclass=Singleton):

    def __init__(self):
        self._hfr_list_data = None

    @property
    def hfr_list_data(self):
        if self._hfr_list_data is None:
            self._hfr_list_data = self.load_hfr_list_data()
        return self._hfr_list_data

    @hfr_list_data.setter
    def hfr_list_data(self, value):
        self._hfr_list_data = value

    @Connector.if_connected(PIPELINE_DATABASE)
    def load_hfr_list_data(self):
        """
        Query ROC database to get the data from hfr_freq_list table.

        :return: query result
        """
        from sqlalchemy import asc, or_
        from roc.dingo.models.packet import TcLog
        # get a database session
        session = Connector.manager[PIPELINE_DATABASE].session
        # Set filters
        logger.debug(f'Connecting {PIPELINE_DATABASE} database to retrieve entries from pipeline.tc_log table ...')
        filters = [TcLog.palisade_id == current_tc
                   for current_tc in TC_HFR_LIST_PAR]
        filters = or_(*filters)

        # Query database (returns results as a pandas.DataFrame object)
        results = pd.read_sql(session.query(TcLog).filter(
            filters).order_by(asc(TcLog.utc_time)).statement, session.bind)

        return results

    def get_freq(self, utc_time, band, mode, get_index=False):
        """
        Get the list of frequencies (kHz) in HFR LIST mode for a given band (HF1 or HF2)

        :param utc_time: datetime object containing the UTC time for which the frequency list must be returned
        :param band: string giving the name of the HFR band (HF1 or HFR2)
        :param mode: string giving the mode (0=NORMAL, 1=BURST)
        :param get_index: If True return the index of the frequencies instead of value
        :return:
        """
        # Get lists of frequencies in HFR LIST mode stored in database
        hfr_list_data = self.hfr_list_data

        # Expected TC name
        band_index = int(band[-1])
        if mode == 1:
            tc_name = f'TC_THR_LOAD_BURST_PAR_{band_index + 1}'
            param_name = f'SY_THR_B_SET_HLS_FREQ_HF{band_index}'
        else:
            tc_name = f'TC_THR_LOAD_NORMAL_PAR_{band_index + 1}'
            param_name = f'SY_THR_N_SET_HLS_FREQ_HF{band_index}'

        # Get list of frequencies from the last TC executed on board
        try:
            latest_hfr_list_data = hfr_list_data[(
                hfr_list_data.palisade_id == tc_name) & (hfr_list_data.utc_time <= utc_time)].iloc[-1]
        except:
            logger.warning(f'No valid frequency list found for {tc_name} at {utc_time}!')
            frequency_list = []
        else:
            # Get indices of frequencies for current utc_time, mode and band
            frequency_list = latest_hfr_list_data.data[param_name].copy()
            logger.debug(f'Frequency values found for {tc_name} on {latest_hfr_list_data.utc_time}: {frequency_list}')

        if not get_index:
            # Compute frequency values in kHz
            frequency_list = [compute_hfr_list_freq(current_freq)
                              for current_freq in frequency_list]

        return frequency_list


def compute_hfr_list_freq(freq_index):
    """
    In HFR LIST mode, return frequency value in kHz giving its
    index

    :param freq_index: index of the frequency
    :return: Value of the frequency in kHz
    """
    return 375 + 50 * (int(freq_index) - 436)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from roc.rap.tasks.tds.normal_rswf import set_rswf
from roc.rap.tasks.tds.normal_tswf import set_tswf

"""
Contains wrapper method to process SBM1 and SBM2 TDS L1 data.
"""

__all__ = ["rswf", "tswf"]

def rswf(l0, task):
    """
    Method to call normal_rswf.set_rswf() with is_sbm input keyword as True.

    :param l0: L0 HDF5 object containing packet data
    :param task: Task class instance
    :return: numpy.ndarray with TDS_SCIENCE_SBM1_RSWF L1 data
    """
    return set_rswf(l0, task, is_sbm=True)


def tswf(l0, task):
    """
    Method to call normal_tswf.set_tswf() with is_sbm input keyword as True.

    :param l0: L0 HDF5 object containing packet data
    :param task: Task class instance
    :return: numpy.ndarray with TDS_SCIENCE_SBM1_TSWF L1 data
    """
    return set_tswf(l0, task, is_sbm=True)


import numpy as np

from poppy.core.logger import logger

from roc.rpl.time import Time
from roc.idb.converters.cdf_fill import fill_records
from libc.stdint cimport uint8_t
from libc.stdint cimport uint16_t

from roc.rap.tasks.utils import sort_data

__all__ = ["PSD", "set_psd"]

# Numpy array dtype for TDS LFM PSD L1 survey data
# Name convention is lower case

PSD = [
    ("epoch", "uint64"),
    ("scet", "float64"),
    ("acquisition_time", ("uint32", 2)),
    ("synchro_flag", "uint8"),
    ("quality_flag", "uint8"),
    ("quality_bitmask", "uint8"),
    ("survey_mode", "uint8"),
    ("bia_status_info", ("uint8", 6)),
    ("lf_data_artefacts", ("uint8", 16)),
    ("input_config", ("uint8", 6)),
    ("channel_status_info", ("uint8", 6)),
    ("psd_srclen", "uint8"),
    ("psd_freq_nr", "uint16"),
    ("psd_freq_axis", "uint8"),
    ("psd_data", ("uint16", (6, 200))),
]


cpdef set_psd(l0, task):
    cdef:
        Py_ssize_t index
        Py_ssize_t channel
        Py_ssize_t size
        uint16_t number
        uint8_t counter

    # name of the packet
    name = "TM_TDS_SCIENCE_LFM_PSD"
    logger.debug("Get data for " + name)

    # get the LFM PDS packet data
    if name in l0["TM"]:
        tm = l0["TM"][name]["source_data"]
    else:
        # rest of the code inside the with statement will not be called
        return np.empty(
            0,
            dtype=PSD,
        )

    # Sort input packets by ascending acquisition times
    tm, __ = sort_data(tm, (
        tm["PA_TDS_ACQUISITION_TIME"][:,1],
        tm["PA_TDS_ACQUISITION_TIME"][:,0],
                    ))

    # get the number of packets
    size = tm["PA_TDS_ACQUISITION_TIME"].shape[0]
    logger.debug("Number of packets: {0}".format(size))

    # create the data array with good size
    data = np.empty(size, dtype=PSD)
    fill_records(data)

    # survey mode to normal
    data["survey_mode"][:] = 0

    # TT2000 Epoch times
    data["epoch"][:] = Time().obt_to_utc(tm["PA_TDS_ACQUISITION_TIME"][:, :2], to_tt2000=True)

    # acquisition time
    data["acquisition_time"][:, :] = tm["PA_TDS_ACQUISITION_TIME"][:, :2]
    data["synchro_flag"][:] = tm["PA_TDS_ACQUISITION_TIME"][:, 2]

    # scet
    data["scet"][:] = Time.cuc_to_scet(data["acquisition_time"])

    # channel status info
    data["channel_status_info"][:, :] = np.vstack(
        (
            tm["PA_TDS_LFM_PSD_CH1"],
            tm["PA_TDS_LFM_PSD_CH2"],
            tm["PA_TDS_LFM_PSD_CH3"],
            tm["PA_TDS_LFM_PSD_CH4"],
            tm["PA_TDS_LFM_PSD_CH5"],
            tm["PA_TDS_LFM_PSD_CH6"],
        )
    ).T

    # artefacts
    data["lf_data_artefacts"][:, :] = np.vstack(
        (
            tm["PA_TDS_LF_ART_CH1_OR"],
            tm["PA_TDS_LF_ART_CH2_OR"],
            tm["PA_TDS_LF_ART_CH3_OR"],
            tm["PA_TDS_LF_ART_CH4_OR"],
            tm["PA_TDS_LF_ART_CH5_OR"],
            tm["PA_TDS_LF_ART_CH6_OR"],
            tm["PA_TDS_LF_ART_CH7_OR"],
            tm["PA_TDS_LF_ART_CH8_OR"],
            tm["PA_TDS_LF_ART_MAG_HEATER"],
            tm["PA_TDS_LF_ART_SCM_CALIB"],
            tm["PA_TDS_LF_ART_ANT3_OFF"],
            tm["PA_TDS_LF_ART_ANT2_OFF"],
            tm["PA_TDS_LF_ART_ANT1_OFF"],
            tm["PA_TDS_LF_ART_SCM_OFF"],
            tm["PA_TDS_LF_ART_THR_OFF"],
            tm["PA_TDS_LF_ART_LFR_OFF"],
        )
    ).T

    # input config
    data["input_config"][:, :] = np.vstack(
        (
            tm["PA_TDS_LFM_PSD_INPUT_CONF1"],
            tm["PA_TDS_LFM_PSD_INPUT_CONF2"],
            tm["PA_TDS_LFM_PSD_INPUT_CONF3"],
            tm["PA_TDS_LFM_PSD_INPUT_CONF4"],
            tm["PA_TDS_LFM_PSD_INPUT_CONF5"],
            tm["PA_TDS_LFM_PSD_INPUT_CONF6"],
        )
    ).T

    # bias status info
    data["bia_status_info"][:, :] = np.vstack(
        (
            tm["PA_BIA_ON_OFF"],
            tm["PA_BIA_MODE_BIAS3_ENABLED"],
            tm["PA_BIA_MODE_BIAS2_ENABLED"],
            tm["PA_BIA_MODE_BIAS1_ENABLED"],
            tm["PA_BIA_MODE_HV_ENABLED"],
            tm["PA_BIA_MODE_MUX_SET"],
        )
    ).T

    # other PSD data
    data["psd_srclen"][:] = tm["PA_TDS_LFM_PSD_SRCLEN"]
    data["psd_freq_axis"][:] = tm["PA_TDS_LFM_PSD_FREQ_AXIS"]
    data["psd_freq_nr"][:] = tm["PA_TDS_LFM_PSD_NUM_FREQS"]

    # loop over the data to assign to the good channel
    with nogil:
        for index in range(size):
            with gil:
                # initialize the counter for channels
                counter = 0

                # number of freqs
                number = data["psd_freq_nr"][index]

                # loop over channels to check which one is activated and set
                # the data according to this
                for channel in range(6):
                    # if the channel is activated
                    if data["channel_status_info"][index, channel] == 1:
                        # copy only used data
                        data["psd_data"][index, channel, :number] = tm[
                            "PA_TDS_LFM_PSD_DATA_BLK"
                        ][index, counter * number:(counter + 1) * number]

                        # increment the counter
                        counter += 1

    return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Exceptions definition for RAP plugin.
"""

from poppy.core.logger import logger

__all__ = ['MetadataException']

class MetadataException(Exception):
    """Exception raise if issue with metadata."""
    def __init__(self, message, *args, **kwargs):
        super(MetadataException, self).__init__(*args, **kwargs)
        logger.error(message)
        self.message = message


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Constants for RAP plugin.
"""

__all__ = ['PLUGIN',
           'PIPELINE_DATABASE',
           'TIME_SQL_STRFORMAT',
           'TC_HFR_LIST_PAR',
           'UINT32_MAX_ENCODING',
           'THR_TICK_NSEC']

PLUGIN = 'roc.rap'

# Load pipeline database identifier
try:
    from poppy.core.logger import logger
    from poppy.core.conf import settings
    PIPELINE_DATABASE = settings.PIPELINE_DATABASE
except Exception:
    PIPELINE_DATABASE = 'PIPELINE_DATABASE'
    logger.warning(f'settings.PIPELINE_DATABASE not defined for {__file__}, \
                     use "{PIPELINE_DATABASE}" by default!')

# SQL string time format
TIME_SQL_STRFORMAT = '%Y-%m-%d %H:%M:%S.%f'

# List of TC for HFR LIST mode configuration
TC_HFR_LIST_PAR = [
                    'TC_THR_LOAD_NORMAL_PAR_2',
                    'TC_THR_LOAD_NORMAL_PAR_3',
                    'TC_THR_LOAD_BURST_PAR_2',
                    'TC_THR_LOAD_BURST_PAR_3',
                   ]

# Max encoding value for uint32
UINT32_MAX_ENCODING = 2**32

# THR Tick value in nanosec
THR_TICK_NSEC = 15258



import numpy as np
cimport numpy as np

from poppy.core.logger import logger
from poppy.core.tools.exceptions import print_exception

from cpython cimport bool
from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t

from roc.rpl.time import Time
from roc.idb.converters.cdf_fill import fill_records, CDFToFill

from roc.rap.tasks.utils import sort_data

__all__ = ["group", "start_end_group", "fill_asm", "set_cwf", "set_swf"]


# Expected number of ASM bins per buffer for F0, F1 and F2
# A ASM buffer is formed by a set of 3 successive packets
ASM_NBINS = {"F0": [32, 32, 24],
             "F1": [36, 36, 32],
             "F2": [32, 32, 32]}

# ASM Bins offset
ASM_OFFSET = {"F0": 17,
              "F1": 6,
              "F2": 7}

cpdef tuple count_buff(np.ndarray[np.uint8_t, ndim=1] x):
    cdef:
        Py_ssize_t i
        Py_ssize_t nmax = x.shape[0]
        np.ndarray buff = np.zeros(nmax, dtype=np.uint32)
        uint32_t nbuff = 0

    # Get the number of buffers
    for i in range(1, nmax):
        if x[i] <= x[i - 1]:
            nbuff += 1
        buff[i] = nbuff

    return buff, nbuff + 1


cpdef tuple group(np.ndarray[np.uint8_t, ndim=1] x):
    cdef:
        Py_ssize_t i
        Py_ssize_t nmax = x.shape[0]
        np.ndarray result = np.zeros(nmax, dtype=np.uint64)
        Py_ssize_t counter = 0

    # Get the number of groups (i.e., buffer)
    for i in range(1, nmax):
        if x[i] <= x[i - 1]:
            counter += 1
        result[i] = counter

    return result, start_end_group(x, counter + 1)


cpdef np.ndarray[np.uint64_t, ndim = 2] start_end_group(
    np.ndarray[np.uint8_t, ndim=1] x,
    uint64_t ngroups,
):
    cdef:
        Py_ssize_t i
        Py_ssize_t nmax = x.shape[0]
        Py_ssize_t counter = 0
        np.ndarray[np.uint64_t, ndim = 2] result = np.zeros(
            (ngroups, 2),
            dtype=np.uint64,
        )
        Py_ssize_t start = 0

    for i in range(1, nmax):
        if x[i] <= x[i - 1]:
            # indicate the start and end of new group
            result[counter, 0] = start
            result[counter, 1] = i - 1

            # new group
            counter += 1

            # new start point
            start = i

    # treat the last group
    result[counter, 0] = start
    result[counter, 1] = nmax - 1

    return result


cpdef np.ndarray fill_asm(dict fdata, str freq, list asm_dtype, np.ndarray asm_blks):

    cdef:
        Py_ssize_t i
        Py_ssize_t j
        Py_ssize_t k
        Py_ssize_t l
        list nbins
        int bin_offset
        int ibuff_len
        np.ndarray ibuff
        np.ndarray[np.uint8_t, ndim = 1] nblk
        uint32_t nbuff
        np.ndarray[np.uint8_t, ndim = 1] nr_asm
        np.ndarray[np.uint32_t, ndim = 1] buff
        np.ndarray data
        np.ndarray[np.float32_t, ndim = 2] asm_buff

    # Make sure that the input packets are sorted by ascending acquisition time
    # and ASM buffer packet index
    fdata, __ = sort_data(fdata, (
        fdata["PA_LFR_PKT_NR_ASM"][:],
        fdata["PA_LFR_ACQUISITION_TIME"][:, 1],
        fdata["PA_LFR_ACQUISITION_TIME"][:, 0],
    ))

    # Packet index for a given ASM buffer
    # In nominal, 1 ASM data buffer is stored into 3 successive packets
    nr_asm = fdata["PA_LFR_PKT_NR_ASM"]

    # Actual number of blocks per packet
    nblk = fdata["PA_LFR_ASM_{0}_BLK_NR".format(freq)].astype(np.uint8)

    # Expected number of bins per buffer
    nbins = ASM_NBINS[freq]

    # Get bin offset for current freq
    bin_offset = ASM_OFFSET[freq]

    # Number of buffers present in the data
    buff, nbuff = count_buff(nr_asm)

    # Create the data container from the number of buffers
    data = np.empty(nbuff, dtype=asm_dtype)
    fill_records(data)

    # Initalize ASM buffer data array
    asm_buff = np.full([128, 25], -1.0e31, dtype=np.float32)

    # Loop on each buffer
    i = 0
    while i < nbuff:

        # Get indices for the current ASM buffer
        # (should have 3 elements
        # corresponding to the 3 successive packets of data)
        ibuff = np.where(buff == i)[0]

        # Get number of packets
        ibuff_len = len(ibuff)
        if ibuff_len != 3:
            logger.warning('Current LFR ASM buffer is not complete!')

        # get the acquisition time and synchro of
        # the first packet available of the buffer
        data["acquisition_time"][i] = fdata[
            "PA_LFR_ACQUISITION_TIME"][ibuff[0], :2]
        data["synchro_flag"][i] = fdata[
            "PA_LFR_ACQUISITION_TIME"][ibuff[0], 2]

        # Compute the corresponding TT2000 epoch times
        data["epoch"][i] = Time().obt_to_utc(fdata[
            "PA_LFR_ACQUISITION_TIME"][ibuff[0], :2], to_tt2000=True)

        data["scet"][i] = Time.cuc_to_scet(fdata[
            "PA_LFR_ACQUISITION_TIME"][ibuff[0], :2])

        # Fill parameters related to LFR config (e.g., Bias)
        data["bias_mode_mux_set"][i] = fdata["PA_BIA_MODE_MUX_SET"][ibuff[0]]
        data["bias_mode_hv_enabled"][i] = fdata[
            "PA_BIA_MODE_HV_ENABLED"][ibuff[0]]
        data["bias_mode_bias1_enabled"][i] = fdata[
            "PA_BIA_MODE_BIAS1_ENABLED"][ibuff[0]]
        data["bias_mode_bias2_enabled"][i] = fdata[
            "PA_BIA_MODE_BIAS2_ENABLED"][ibuff[0]]
        data["bias_mode_bias3_enabled"][i] = fdata[
            "PA_BIA_MODE_BIAS3_ENABLED"][ibuff[0]]
        data["bias_on_off"][i] = fdata["PA_BIA_ON_OFF"][ibuff[0]]
        data["bw"][i] = fdata["SY_LFR_BW"][ibuff[0]]
        data["sp0"][i] = fdata["SY_LFR_SP0"][ibuff[0]]
        data["sp1"][i] = fdata["SY_LFR_SP1"][ibuff[0]]
        data["r0"][i] = fdata["SY_LFR_R0"][ibuff[0]]
        data["r1"][i] = fdata["SY_LFR_R1"][ibuff[0]]
        data["r2"][i] = fdata["SY_LFR_R2"][ibuff[0]]

        # Filling ASM over the 3 successive packets of the buffer
        asm_buff[:] = -1.0e31

        # Define initial offset for the current buffer
        k = bin_offset
        l = 0
        # Loop on ASM expected packets for current buffer (i.e., 3)
        for j in range(len(nbins)):

            if j + 1 != nr_asm[ibuff[l]]:
                logger.warning(f"Packet #{j} is missing!")
                continue
            else:
                # If the number of blocks in the packet is not
                # equal to the expected number of bins
                # thus the current packet is not complete
                if nbins[j] != nblk[ibuff[l]]:
                    logger.warning("Incomplete LFR ASM packet!")

                # Fill asm_buff array with ASM blocks for the l-th packet in
                # the current buffer
                asm_buff[k:k + nblk[ibuff[l]],
                         :] = asm_blks[ibuff[l], :nblk[ibuff[l]], :]

                # Increment local ibuff indice
                l += 1
                if l >= ibuff_len:
                    break

            # Increment by bins length
            k += nbins[j]

        # Fill output data array with current ASM buffer data
        data["asm"][i, :, :] = asm_buff[:, :]

        # Increment buffer index
        i += 1

    data["asm_cnt"][:] = np.sum(nbins)

    return data


# LFR CWF SAMPLING RATES IN Hz
CWF_SAMPLING_RATE = {"F1": 4096.0,
                     "F2": 256.0,
                     "F3": 16.0}

# Fill CDF data with CWF packet data
cpdef np.ndarray set_cwf(dict fdata, str freq, list cwf_dtype, bool magnetic, task):

    cdef:
        Py_ssize_t packet
        uint16_t nsamp
        uint64_t nsamps, size, i
        uint64_t epoch0
        float delta_t
        float samp_rate
        np.ndarray blk_nr
        np.ndarray data
        np.ndarray elec
        np.ndarray mag

    # Sort input packets by ascending acquisition times
    fdata, __ = sort_data(fdata, (
        fdata["PA_LFR_ACQUISITION_TIME"][:, 1],
        fdata["PA_LFR_ACQUISITION_TIME"][:, 0],
    ))

    # Get number of blocks
    if freq == "F3":
        blk_nr = fdata["PA_LFR_CWF3_BLK_NR"][:]
    elif freq == "LONG_F3":
        blk_nr = fdata["PA_LFR_CWFL3_BLK_NR"][:]
        freq = freq[-2:]
    else:
        blk_nr = fdata["PA_LFR_CWF_BLK_NR"][:]

    # Number of samples
    nsamps = np.sum(blk_nr)

    # Sampling rate
    samp_rate = CWF_SAMPLING_RATE[freq]

    # Cadence in nanosec
    delta_t = 1.0e9 / samp_rate

    # create a record array with the good dtype and good length
    data = np.empty(nsamps, dtype=cwf_dtype)
    fill_records(data)

    # Get the number of packets
    size = fdata["PA_LFR_ACQUISITION_TIME"].shape[0]

    # Loop over packets
    i = 0
    for packet in range(size):

        # Get the number of samples for the current packet
        nsamp = blk_nr[packet]

        # get the acquisition time and time synch. flag
        data["acquisition_time"][i:i + nsamp, :] = \
            fdata["PA_LFR_ACQUISITION_TIME"][packet, :2]
        data["synchro_flag"][i:i + nsamp] = \
            fdata["PA_LFR_ACQUISITION_TIME"][packet, 2]

        # Compute corresponding Epoch time
        epoch0 = Time().obt_to_utc(
            fdata["PA_LFR_ACQUISITION_TIME"][packet, :2].reshape([1, 2]),
            to_tt2000=True)

        tt = Time().obt_to_utc(
            fdata["PA_LFR_ACQUISITION_TIME"][packet, :2].reshape([1, 2]),
            to_datetime=True)

        data["epoch"][i:i + nsamp] = epoch0 + \
            np.uint64(delta_t * np.arange(0, nsamp, 1))

        data["scet"][
            i:i + nsamp] = Time.cuc_to_scet(data["acquisition_time"][i:i + nsamp, :])

        data["sampling_rate"][i:i + nsamp] = samp_rate

        data["bias_mode_mux_set"][i:i + nsamp] = \
            fdata["PA_BIA_MODE_MUX_SET"][packet]
        data["bias_mode_hv_enabled"][i:i + nsamp] = \
            fdata["PA_BIA_MODE_HV_ENABLED"][packet]
        data["bias_mode_bias1_enabled"][i:i + nsamp] = \
            fdata["PA_BIA_MODE_BIAS1_ENABLED"][packet]
        data["bias_mode_bias2_enabled"][i:i + nsamp] = \
            fdata["PA_BIA_MODE_BIAS2_ENABLED"][packet]
        data["bias_mode_bias3_enabled"][i:i + nsamp] = \
            fdata["PA_BIA_MODE_BIAS3_ENABLED"][packet]
        data["bias_on_off"][i:i + nsamp] = fdata["PA_BIA_ON_OFF"][packet]
        data["bw"][i:i + nsamp] = fdata["SY_LFR_BW"][packet]
        data["sp0"][i:i + nsamp] = fdata["SY_LFR_SP0"][packet]
        data["sp1"][i:i + nsamp] = fdata["SY_LFR_SP1"][packet]
        data["r0"][i:i + nsamp] = fdata["SY_LFR_R0"][packet]
        data["r1"][i:i + nsamp] = fdata["SY_LFR_R1"][packet]
        data["r2"][i:i + nsamp] = fdata["SY_LFR_R2"][packet]

        # potential just need to be reshaped
        data["v"][i:i + nsamp] = fdata["PA_LFR_SC_V_" + freq][packet, :]

        # electrical and magnetic needs to have their components merged before
        # the reshaping
        elec = np.empty([nsamp, 2], dtype=np.int16)
        elec[:, 0] = fdata["PA_LFR_SC_E1_" + freq][packet, :]
        elec[:, 1] = fdata["PA_LFR_SC_E2_" + freq][packet, :]
        data["e"][i:i + nsamp, :] = elec

        # get magnetic only if present
        if magnetic:
            mag = np.empty([nsamp, 3], dtype=np.int16)
            mag[:, 0] = fdata["PA_LFR_SC_B1_" + freq][packet, :]
            mag[:, 1] = fdata["PA_LFR_SC_B2_" + freq][packet, :]
            mag[:, 2] = fdata["PA_LFR_SC_B3_" + freq][packet, :]
            data["b"][i:i + nsamp, :] = mag

        i += nsamp

    return data

# LFR SWF SAMPLING RATES IN Hz
SWF_SAMPLING_RATE = {"F0": 24576.0,
                     "F1": 4096.0,
                     "F2": 256.0}

# Max. Number of WF samples in a packet
NSAMP_MAX_PER_PKT = 304

# Number of samples per snapshot
SAMP_PER_SNAPSHOT = 2048

# Number of packets per snapshot
PKT_PER_SNAPSHOT = 7

# Fill CDF data with SWF packet data
cpdef set_swf(dict fdata, str freq, list swf_dtype, task):
    cdef:
        Py_ssize_t i, j, k
        np.uint16_t samp_per_pkt
        np.uint16_t size
        np.uint16_t snapshot_nsamp
        np.uint16_t samp_per_pkt_miss
        np.uint16_t samp_per_pkt_max
        np.ndarray data
        np.ndarray[np.uint8_t, ndim = 1] pkt_nr
        np.ndarray[np.uint8_t, ndim = 1] pkt_cnt
        np.ndarray[np.uint32_t, ndim = 1] snapshot_id
        np.ndarray[np.uint32_t, ndim = 1] snpht
        np.ndarray[np.uint32_t, ndim = 1] snapshot_seq_nr

    # Sort input packets by ascending acquisition times and snapshot number
    fdata, __ = sort_data(fdata, (
        fdata["PA_LFR_PKT_NR"],
        fdata["PA_LFR_ACQUISITION_TIME"][:, 1],
        fdata["PA_LFR_ACQUISITION_TIME"][:, 0],
    ))

    # Get Total count of packets for the current snapshot.
    pkt_cnt = fdata["PA_LFR_PKT_CNT"]
    # Get Number of the packet for the current snapshot
    pkt_nr = fdata["PA_LFR_PKT_NR"]

    # Get Number of snapshots
    snapshot_seq_nr, size = count_buff(pkt_nr)
    snapshot_id = np.unique(snapshot_seq_nr).astype(np.uint32)

    # create the data array with good size
    data = np.empty(size, dtype=swf_dtype)
    fill_records(data)

    # Max sample number per record for V, E and B
    nsamp_max_per_rec = data["e"].shape[-1]

    # Loop over snapshots
    logger.debug("Number of snapshots: {0}".format(size))
    with nogil:
        for i in range(size):
            with gil:
                snpht = np.where(snapshot_seq_nr == snapshot_id[i])[
                    0].astype(np.uint32)

                # Sort snapshot packets by increasing PA_LFR_PKT_NR values
                snpht = snpht[np.argsort(pkt_nr[snpht[:]])]

                # Fill acquisition time
                data["acquisition_time"][i, :] = fdata[
                    "PA_LFR_ACQUISITION_TIME"][snpht[0], :2]

                data["epoch"][i] = Time().obt_to_utc(data["acquisition_time"][i, :].reshape([1, 2]),
                                                     to_tt2000=True)

                data["scet"][i] = Time.cuc_to_scet(
                    data["acquisition_time"][i, :].reshape([1, 2]))

                # the flag for time synchronization
                data["synchro_flag"][i] = fdata[
                    "PA_LFR_ACQUISITION_TIME"][snpht[0], 2]

                # Sampling rate in Hz
                data["sampling_rate"][i] = float(SWF_SAMPLING_RATE[freq])

                data["bias_mode_mux_set"][i] = fdata[
                    "PA_BIA_MODE_MUX_SET"][snpht[0]]
                data["bias_mode_hv_enabled"][i] = fdata[
                    "PA_BIA_MODE_HV_ENABLED"
                ][snpht[0]]
                data["bias_mode_bias1_enabled"][i] = fdata[
                    "PA_BIA_MODE_BIAS1_ENABLED"
                ][snpht[0]]
                data["bias_mode_bias2_enabled"][i] = fdata[
                    "PA_BIA_MODE_BIAS2_ENABLED"
                ][snpht[0]]
                data["bias_mode_bias3_enabled"][i] = fdata[
                    "PA_BIA_MODE_BIAS3_ENABLED"
                ][snpht[0]]
                data["bias_on_off"][i] = fdata["PA_BIA_ON_OFF"][snpht[0]]
                data["bw"][i] = fdata["SY_LFR_BW"][snpht[0]]
                data["sp0"][i] = fdata["SY_LFR_SP0"][snpht[0]]
                data["sp1"][i] = fdata["SY_LFR_SP1"][snpht[0]]
                data["r0"][i] = fdata["SY_LFR_R0"][snpht[0]]
                data["r1"][i] = fdata["SY_LFR_R1"][snpht[0]]
                data["r2"][i] = fdata["SY_LFR_R2"][snpht[0]]

                # Max number of samples for this snapshot
                samp_per_pkt_max = NSAMP_MAX_PER_PKT
                samp_per_pkt_miss = samp_per_pkt_max

                # Loop on each packet of the current snapshot
                # Assuming there is no missing packet in the snapshot
                # (i.e., pkt_cnt = 7)
                snapshot_nsamp = 0
                j = 0
                k = 0
                while k < pkt_cnt[snpht[0]]:

                    if j >= len(snpht):
                        j = len(snpht) - 1

                    # Check that packet counter k value
                    # is the same than the packet sequence
                    # number pkt_nr[snpht[j]] in the current snapshot
                    if k + 1 != pkt_nr[snpht[j]]:
                        # ... If not, then there is a missing packet.
                        logger.warning("Packet #{0} [{2}] is missing for snapshot #{1}!".format(
                            k + 1, i, pkt_cnt[snpht[0]]))

                        # If the missing packet is the last packet...
                        if k + 1 == pkt_cnt[snpht[0]]:
                            # compute the number of sample
                            # in the last packet
                            samp_per_pkt_miss = SAMP_PER_SNAPSHOT - \
                                (samp_per_pkt_max * (pkt_cnt[snpht[0]] - 1))

                        # If the missing packet is the first packet...
                        elif k + 1 == 1:
                            # If there is only 2 packets for this snapshot
                            if pkt_cnt[snpht[0]] == 2:
                                # Compute the number of sample
                                # in the first packet
                                samp_per_pkt_miss = SAMP_PER_SNAPSHOT - samp_per_pkt_max
                            else:
                                # ... Get the number of samples from the next
                                # valid packet (which should be
                                # not the last one)
                                samp_per_pkt_miss = samp_per_pkt_max

                        # Increment total number of samples
                        # for the current snapshot
                        # assuming that samp_per_pkt_miss samples
                        # are missing
                        snapshot_nsamp += samp_per_pkt_miss

                        # Increment packet counter k and continue
                        # in order to reach the next packet
                        # sequence number pkt_nr[snpht[j]] in the
                        # snapshot
                        k += 1
                        continue

                    # Get number of LFR SWF sample data blocks
                    samp_per_pkt = fdata["PA_LFR_SWF_BLK_NR"][snpht[j]]

                    # Update the samp_per_pkt_miss value
                    samp_per_pkt_miss = samp_per_pkt

                    # Fill potential values
                    data["v"][i, snapshot_nsamp: snapshot_nsamp +
                              samp_per_pkt] = fdata["PA_LFR_SC_V_" + freq][snpht[j], :samp_per_pkt]

                    # Fill electric field values
                    data["e"][i, 0, snapshot_nsamp: snapshot_nsamp +
                              samp_per_pkt] = fdata["PA_LFR_SC_E1_" + freq][snpht[j], :samp_per_pkt]
                    data["e"][i, 1, snapshot_nsamp: snapshot_nsamp +
                              samp_per_pkt] = fdata["PA_LFR_SC_E2_" + freq][snpht[j], :samp_per_pkt]

                    # Fill magnetic field values
                    data["b"][i, 0, snapshot_nsamp: snapshot_nsamp +
                              samp_per_pkt] = fdata["PA_LFR_SC_B1_" + freq][snpht[j], :samp_per_pkt]
                    data["b"][i, 1, snapshot_nsamp: snapshot_nsamp +
                              samp_per_pkt] = fdata["PA_LFR_SC_B2_" + freq][snpht[j], :samp_per_pkt]
                    data["b"][i, 2, snapshot_nsamp: snapshot_nsamp +
                              samp_per_pkt] = fdata["PA_LFR_SC_B3_" + freq][snpht[j], :samp_per_pkt]

                    # Increment total number of samples per channel
                    # for the current snapshot
                    snapshot_nsamp += samp_per_pkt

                    k += 1
                    j += 1

                # Perform some verifications before saving the snapshot
                try:
                    if snapshot_nsamp > nsamp_max_per_rec:
                        raise Exception("Total number of samples cannot exceed {1}, but {0} is found!".format(
                            snapshot_nsamp, nsamp_max_per_rec))

                    if SAMP_PER_SNAPSHOT != snapshot_nsamp:
                        raise Exception(
                            "Unexpected total number of samples in snapshot #{0}!".format(snapshot_id[i]))
                except:
                    message = print_exception()
                    task.exception(message)
                    data["v"][i, ] = CDFToFill()(data["v"].dtype.name)
                    data["e"][i, ] = CDFToFill()(data["e"].dtype.name)
                    data["b"][i, ] = CDFToFill()(data["b"].dtype.name)

    return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.command import Command

class RapCommands(Command):
    """
    Manage the commands relative to the FILM module.
    """
    __command__ = "rap"
    __command_name__ = "rap"
    __parent__ = "master"
    __parent_arguments__ = ["base"]
    __help__ = """
        Commands relative to the RAP module, responsible for 
        processing RPW L0, L1 and HK data.
    """





import numpy as np
cimport numpy as np

from libc.stdint cimport uint64_t

from poppy.core.tools.exceptions import print_exception

from roc.rpl.time import Time
from roc.rpl.packet_structure.data cimport Data
from roc.rpl.packet_structure.data import Data

from roc.idb.converters.cdf_fill import fill_records as set_fillval

from roc.rap.tasks.thr.science_tools import input_setup
from roc.rap.tasks.thr.science_tools import voltages
from roc.rap.tasks.thr.science_tools import do_analysis
from roc.rap.tasks.thr.science_tools import voltages_dtype
from roc.rap.tasks.thr.science_tools import input_setup_dtype
from roc.rap.tasks.thr.science_tools import do_analysis_dtype
from roc.rap.tasks.thr.tnr import extract_calibration_band as extract_band
from roc.rap.tasks.thr.tnr import tnr_setup
from roc.rap.tasks.thr.tnr import tnr_dtype


class TNRCalibrationDecomError(Exception):
    """
    Error management for TNR calibration.
    """


cpdef decommute(group, task):
    cdef:
        Py_ssize_t i

    # get the size for the number of packets
    number = group["PA_THR_TNR_DATA_CNT"].shape[0]

    # create the array containing the data
    array = np.empty(number, dtype=create_dtype())
    array.fill(np.nan)

    # loop over packets
    for i in range(number):
        try:
            # size for number of unsigned int
            size = group["PA_THR_TNR_DATA_CNT"][i]

            # transform the data into byte array
            # store because of the garbage collector removing reference while
            # processing in cython
            byte = group[
                "PA_THR_TNR_DATA"
            ][i][:size].byteswap().newbyteorder().tobytes()
            data = Data(byte, len(byte))

            # now extract the data
            result, offset = extract_data(
                data,
                0,
                group["PA_THR_ACQUISITION_TIME"][i],
            )

            # check offset
            if offset != size * 2:
                raise TNRCalibrationDecomError(
                    "Expected a size of {0} bytes but have {1}".format(
                        size * 2,
                        offset,
                    )
                )

            # store the data
            array[i] = result
        except:
            # do nothing with this packet and store the error in the database
            message = print_exception()
            task.exception(message)

    # get the number of records
    number, bands = band_record_number(array)

    # create an array of tnr records
    records = np.empty(number, dtype=tnr_dtype())
    set_fillval(records)

    # fill the array with the records
    records = fill_records(array, records)

    return records


cdef fill_records(array, records):
    cdef:
        Py_ssize_t i
        unsigned long nblocks
        unsigned long counter

    # get number of blocks
    nblocks = array.shape[0]

    # init the counter for placing data
    counter = 0

    # loop over fields
    for i in range(nblocks):
        if array["tnr_setup"]["a"][i] == 1:
            counter = set_record(records, array, counter, i, "a", 0)
        if array["tnr_setup"]["b"][i] == 1:
            counter = set_record(records, array, counter, i, "b", 1)
        if array["tnr_setup"]["c"][i] == 1:
            counter = set_record(records, array, counter, i, "c", 2)
        if array["tnr_setup"]["d"][i] == 1:
            counter = set_record(records, array, counter, i, "d", 3)

    return records


cdef unsigned long set_record(records, array, counter, i, band, index):
    cdef:
        double epoch
        Py_ssize_t j
        str level

    epoch = <double>Time().obt_to_utc(
                np.array([array["acquisition_time"][i][:2]]),
        to_tt2000=True)

    for j in range(9):
        level = str(j)
        records["epoch"][counter] = epoch
        records["acquisition_time"][counter] = array["acquisition_time"][i][:2]
        records["scet"][counter] = Time.cuc_to_scet(
            np.array([array["acquisition_time"][i][:2]]))
        records["synchro_flag"][counter] = array[
            "acquisition_time"
        ][i][2]
        records["survey_mode"][counter] = 2
        records["average_nr"][counter] = array["tnr_setup"]["av"][i]
        records["auto_cross_status"][counter] = (
            array["tnr_setup"]["au"][i],
            array["tnr_setup"]["cr"][i],
        )
        records["channel_status"][counter] = (
            array["do_analysis"]["ch1"][i],
            array["do_analysis"]["ch2"][i],
        )
        records["calibration_level"][counter] = j
        records["front_end"][counter] = array["input_setup"]["fe"][i]
        records["sensor_config"][counter] = (
            array["input_setup"]["sensor_tnr_1"][i],
            array["input_setup"]["sensor_tnr_2"][i],
        )

        # in calibration mode, no temperatures array, but just PCB temperture
        # and voltages
        records["temperature"][counter] = (
            array["temperature_pcb"][i],
            array["voltages"]["minus5"][i],
            array["voltages"]["plus5"][i],
            array["voltages"]["plus12"][i],
        )
        records["tnr_band"][counter] = index
        records["agc1"][counter] = array["data"][level][band]["ch1"]["agc"][i]
        records["agc2"][counter] = array["data"][level][band]["ch2"]["agc"][i]
        records["auto1"][counter][:8] = array["data"][
            level
        ][band]["ch1"]["autos"][i]
        records["auto2"][counter][:8] = array["data"][
            level
        ][band]["ch2"]["autos"][i]
        records["cross_r"][counter][:8] = array["data"][
            level
        ][band]["cross"][i][:, 0]
        records["cross_i"][counter][:8] = array["data"][
            level
        ][band]["cross"][i][:, 1]
        counter += 1

    return counter

cdef tuple band_record_number(array):
    """
    Get the number of records for each band.
    """
    cdef:
        tuple nnormal
        unsigned long number = 0

    nnormal = (
        int(array["tnr_setup"]["a"].sum()) * 9,
        int(array["tnr_setup"]["b"].sum()) * 9,
        int(array["tnr_setup"]["c"].sum()) * 9,
        int(array["tnr_setup"]["d"].sum()) * 9,
    )
    number += sum(nnormal)

    return number, nnormal



cdef tuple extract_data(Data data, Py_ssize_t offset, time):
    """
    Extract the data and insert it into the data structure of the packet.
    """
    cdef:
        unsigned short cmd_ctr
        unsigned short bit_count
        tuple voltages_cmd
        tuple tnr_setup_cmd
        tuple input_setup_cmd
        tuple do_analysis_cmd
        tuple res
        list banda, bandb, bandc, bandd
        Py_ssize_t i

    # read command control
    cmd_ctr = data.u8(offset, 0, 8)
    offset += 1

    # read temperatures
    voltages_cmd = voltages(data, offset)
    offset += 3

    # read tnr setup
    offset += 1
    tnr_setup_cmd = tnr_setup(data, offset)
    offset += 1

    # read input setup
    input_setup_cmd = input_setup(data, offset)
    offset += 2

    # read do analysis command
    offset += 1
    do_analysis_cmd = do_analysis(data, offset)
    offset += 1

    # initialize bit count
    bit_count = 0

    # init structure for calibration levels
    banda, bandb, bandc, bandd = [], [], [], []

    # loop over the calibration levels
    for i in range(9):
        # extract bands
        res, bit_count = extract_band(
            tnr_setup_cmd[6],
            do_analysis_cmd[4],
            do_analysis_cmd[3],
            tnr_setup_cmd[2],
            tnr_setup_cmd[1],
            data,
            offset,
            bit_count,
        )
        banda.append(res)
        res, bit_count = extract_band(
            tnr_setup_cmd[5],
            do_analysis_cmd[4],
            do_analysis_cmd[3],
            tnr_setup_cmd[2],
            tnr_setup_cmd[1],
            data,
            offset,
            bit_count,
        )
        bandb.append(res)
        res, bit_count = extract_band(
            tnr_setup_cmd[4],
            do_analysis_cmd[4],
            do_analysis_cmd[3],
            tnr_setup_cmd[2],
            tnr_setup_cmd[1],
            data,
            offset,
            bit_count,
        )
        bandc.append(res)
        res, bit_count = extract_band(
            tnr_setup_cmd[3],
            do_analysis_cmd[4],
            do_analysis_cmd[3],
            tnr_setup_cmd[2],
            tnr_setup_cmd[1],
            data,
            offset,
            bit_count,
        )
        bandd.append(res)

    # compute the new offset for the block
    offset += bit_count // 8

    # align
    if offset % 2 != 0:
        offset += 1

    # tuple of values to add them to the numpy array
    return (
        cmd_ctr, voltages_cmd, tnr_setup_cmd,
        input_setup_cmd, do_analysis_cmd,
        tuple([(banda[i], bandb[i], bandc[i], bandd[i]) for i in range(9)]),
        time
    ), offset


cpdef list create_dtype():
    """
    Create dtype of numpy for the special packet.
    """
    return [
        ("temperature_pcb", "uint8"),
        ("voltages", voltages_dtype()),
        ("tnr_setup", tnr_setup_dtype()),
        ("input_setup", input_setup_dtype()),
        ("do_analysis", do_analysis_dtype()),
        ("data", block_data_dtype()),
        ("acquisition_time", "uint32", 3),
    ]


cpdef list tnr_setup_dtype():
    """
    Return the TNR setup command dtype.
    """
    return [
        ("av", "uint8"), ("cr", "uint8"), ("au", "uint8"),
        ("d", "uint8"), ("c", "uint8"), ("b", "uint8"), ("a", "uint8"),
    ]


cpdef list channel_dtype():
    return [
        ("agc", "uint16"), ("autos", "uint16", 8),
    ]


cpdef list band_dtype():
    return [
        ("ch1", channel_dtype()),
        ("ch2", channel_dtype()),
        ("cross", "uint16", (8, 2)),
    ]


cpdef list block_data_dtype():
    return [
        (
            "{0}".format(i),
            [
                ("a", band_dtype()),
                ("b", band_dtype()),
                ("c", band_dtype()),
                ("d", band_dtype())
            ],
        ) for i in range(9)
    ]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
cimport numpy as np

from roc.rpl.time import Time
from roc.idb.converters.cdf_fill import fill_records as filler
from poppy.core.tools.exceptions import print_exception
from roc.rpl.packet_structure.data import Data
from roc.rpl.packet_structure.data cimport Data
from libc.stdint cimport uint64_t
from roc.rap.tasks.thr.science_tools import rpw_status
from roc.rap.tasks.thr.science_tools import temperatures_command
from roc.rap.tasks.thr.science_tools import input_setup
from roc.rap.tasks.thr.science_tools import do_analysis
from roc.rap.tasks.thr.science_tools import rpw_status_dtype
from roc.rap.tasks.thr.science_tools import temperatures_dtype
from roc.rap.tasks.thr.science_tools import input_setup_dtype
from roc.rap.tasks.thr.science_tools import do_analysis_dtype
from roc.rap.tasks.thr.science_tools import average_nr
from roc.rap.tasks.thr.tnr import delta_times
from roc.rap.tasks.thr.tnr import extract_band
from roc.rap.tasks.thr.tnr import tnr_setup
from roc.rap.tasks.thr.tnr import tnr_dtype
from roc.rap.tasks.thr.tnr import DURATION

# Max encoding value for uint32
UINT32_MAX_ENCODING = 2**32

# THR Tick value in nanosec
THR_TICK_NSEC = 15258

# List of TNR frequency values in Hz
TNR_FREQUENCY_LIST = 3991.785737 * np.power(4, 1./32)**np.arange(0, 128)

# List of TNR frequency for each band (0=A, 1=B, 2=C and 3=D)
# List is given as a dictionary with band index as keyword
TNR_BAND_FREQUENCY = {
             0:TNR_FREQUENCY_LIST[0:32],
             1:TNR_FREQUENCY_LIST[32:64],
             2:TNR_FREQUENCY_LIST[64:96],
             3:TNR_FREQUENCY_LIST[96:128],
            }


cpdef decommute_burst(group, task):
    return decommute(group, task, "PA_THR_B_BLOCK_CNT", "burst")


cpdef decommute_normal(group, task):
    return decommute(group, task, "PA_THR_N_BLOCK_CNT", "normal")


cdef decommute(group, task, cnt_param, message):
    cdef:
        Py_ssize_t i
        Py_ssize_t j
        Py_ssize_t counter
        unsigned short size
        unsigned short nblocks

    # get the size for the number of blocks
    number = np.sum(group[cnt_param])

    # create the array containing the data
    array = np.empty(number, dtype=create_dtype())
    array.fill(np.nan)

    # the number of packets
    packet_number = group[cnt_param].shape[0]

    # loop over packets
    counter = 0
    for i in range(packet_number):
        try:
            # size of the number of unsigned int
            size = group["PA_THR_TNR_DATA_CNT"][i]

            # number of blocks
            nblocks = group[cnt_param][i]

            # transform the data into byte array
            # store because of the garbage collector removing reference while
            # processing in Cython
            byte = group[
                "PA_THR_TNR_DATA"
            ][i][:size].byteswap().newbyteorder().tobytes()
            data = Data(byte, len(byte))

            # init offset
            offset = 0

            # loop over blocks
            for j in range(nblocks):
                # now extract the data
                array[counter], offset = extract_data(
                    data,
                    offset,
                    group["PA_THR_ACQUISITION_TIME"][i],
                )

                # increment counter
                counter += 1

            # check offset size
            if offset != size * 2:
                raise Exception(
                    "Bad decommutation for TNR " + message + " mode\n" +
                    "Expected {0} got {1} bytes".format(size * 2, offset)
                )
        except:
            # continue if something wrong with the packet
            message = print_exception()
            task.exception(message)

    # get the number of records
    number, bands = band_record_number(array)

    # create an array of tnr records
    records = np.empty(number, dtype=tnr_dtype())
    filler(records)

    # fill the array with the records
    records = fill_records(array, records, message)

    return records


cdef fill_records(array, records, message):
    cdef:
        Py_ssize_t i
        long nblocks
        unsigned long counter
        np.ndarray[np.uint64_t, ndim=1] delta_offset

    # get number of blocks
    nblocks = array.shape[0]

    # select the mode
    if message == "normal":
        mode = 0
    else:
        mode = 1

    # init the counter for placing data
    counter = 0

    # Init delta time offset for band A, B, C and D
    # (Required of delta times exceed max 32 bits encoding value in packet)
    delta_offset = np.zeros(4, dtype=np.uint64)

    # loop over fields
    for i in range(nblocks):
        if array["tnr_setup"]["a"][i] == 1:
            delta_offset = set_record(records, array, mode,
                                      counter, i, "a", 0,
                                      delta_offset)
            counter += 1
        if array["tnr_setup"]["b"][i] == 1:
            delta_offset = set_record(records, array, mode,
                                      counter, i, "b", 1,
                                      delta_offset)
            counter += 1
        if array["tnr_setup"]["c"][i] == 1:
            delta_offset = set_record(records, array, mode,
                                      counter, i, "c", 2,
                                      delta_offset)
            counter += 1
        if array["tnr_setup"]["d"][i] == 1:
            delta_offset = set_record(records, array, mode,
                                      counter, i, "d", 3,
                                      delta_offset)
            counter += 1

    return records


cdef np.ndarray set_record(
    records,
    array,
    unsigned char mode,
    unsigned long counter,
    Py_ssize_t i,
    str band,
    unsigned char index,
    np.ndarray delta_offset
):
    cdef:
        long delta_time
        long measurement_duration

    #  # Check if UINT32_MAX_ENCODING is reached
    # TODO - This part does not work because of the
    #        issue https://gitlab.obspm.fr/ROC/RCS/THR_CALBAR/-/issues/45
    # is_true = (i > 0 and
    #         Time.cuc_to_scet(
    #         np.array(array["acquisition_time"][i-1][:2])) == Time.cuc_to_scet(
    #         np.array(array["acquisition_time"][i][:2])) and
    #         array["delta_times"][band][i-1] > array["delta_times"][band][i])
    #
    # if is_true:
    #     delta_offset[index] += UINT32_MAX_ENCODING

    delta_time = ((array["delta_times"][band][i] +
                  delta_offset[index]) * THR_TICK_NSEC) # in nanosec
    measurement_duration = DURATION[
        (array["tnr_setup"]["av"][i], index)
    ]

    records["epoch"][counter] = (<long>Time().obt_to_utc(
        np.array([array["acquisition_time"][i][:2]]), to_tt2000=True) + # acquisition time
        delta_time + # delta time in nanoseconds
        0.5 * measurement_duration * 1000000000.0) # duration measure

    records["scet"][counter] = (Time.cuc_to_scet(
        np.array([array["acquisition_time"][i][:2]])) + # acquisition time
        float(delta_time / 1000000000.) + # delta time in seconds
        float(0.5 * measurement_duration) # duration measure
    )
    records["sweep_num"][counter] = i
    records["acquisition_time"][counter] = array["acquisition_time"][i][:2]
    records["synchro_flag"][counter] = array[
        "acquisition_time"
    ][i][2]
    records["measurement_duration"][counter] = measurement_duration * 1000000.
    records["ticks_nr"][counter] = array["delta_times"][band][i] + delta_offset[index]

    records["delta_time"][counter] = delta_time / 1000. # in microseconds
    records["survey_mode"][counter] = mode
    records["average_nr"][counter] = average_nr(array["tnr_setup"]["av"][i])
    records["auto_cross_status"][counter] = (
        array["tnr_setup"]["au"][i],
        array["tnr_setup"]["cr"][i],
    )
    records["channel_status"][counter] = (
        array["do_analysis"]["ch1"][i],
        array["do_analysis"]["ch2"][i],
    )
    records["front_end"][counter] = array["input_setup"]["fe"][i]
    records["sensor_config"][counter] = (
        array["input_setup"]["sensor_tnr_1"][i],
        array["input_setup"]["sensor_tnr_2"][i],
    )
    records["rpw_status"][counter] = array["rpw_status"][i].tolist()
    records["temperature"][counter] = array["temperatures"][i].tolist()
    records["tnr_band"][counter] = index
    records["frequency"][counter] = TNR_BAND_FREQUENCY[index]
    records["agc1"][counter] = array["data"][band]["ch1"]["agc"][i]
    records["agc2"][counter] = array["data"][band]["ch2"]["agc"][i]
    records["auto1"][counter] = array["data"][band]["ch1"]["autos"][i]
    records["auto2"][counter] = array["data"][band]["ch2"]["autos"][i]
    records["cross_r"][counter] = array["data"][band]["cross"][i][:, 0]
    records["cross_i"][counter] = array["data"][band]["cross"][i][:, 1]

    return delta_offset

cdef tuple band_record_number(array):
    """
    Get the number of records for each band.
    """
    cdef:
        tuple nnormal
        unsigned long number = 0

    nnormal = (
        int(array["tnr_setup"]["a"].sum()),
        int(array["tnr_setup"]["b"].sum()),
        int(array["tnr_setup"]["c"].sum()),
        int(array["tnr_setup"]["d"].sum()),
    )
    number += sum(nnormal)

    return number, nnormal


cdef tuple extract_data(Data data, Py_ssize_t offset, time):
    """
    Extract the data and insert it into the data structure of the packet.
    """
    cdef:
        Py_ssize_t i
        unsigned short bit_count
        unsigned short cmd_ctr
        tuple rpw_status_cmd
        tuple temperatures
        tuple tnr_setup_cmd
        tuple input_setup_cmd
        tuple do_analysis_cmd
        tuple deltas
        tuple banda, bandb, bandc, bandd

    # read command control
    cmd_ctr = data.u16(offset, 0, 16)
    offset += 2

    # read rpw status
    rpw_status_cmd = rpw_status(data, offset)
    offset += 2

    # read temperatures
    temperatures = temperatures_command(data, offset)
    offset += 4

    # read tnr setup
    offset += 1
    tnr_setup_cmd = tnr_setup(data, offset)
    offset += 1


    # read input setup
    input_setup_cmd = input_setup(data, offset)
    offset += 2

    # read do analysis command
    offset += 1
    do_analysis_cmd = do_analysis(data, offset)
    offset += 1

    # read delta times
    deltas = delta_times(data, offset)
    offset += 16

    # initialize bit count
    bit_count = 0

    # extract bands
    banda, bit_count = extract_band(
        tnr_setup_cmd[6],
        do_analysis_cmd[4],
        do_analysis_cmd[3],
        tnr_setup_cmd[2],
        tnr_setup_cmd[1],
        data,
        offset,
        bit_count,
    )
    bandb, bit_count = extract_band(
        tnr_setup_cmd[5],
        do_analysis_cmd[4],
        do_analysis_cmd[3],
        tnr_setup_cmd[2],
        tnr_setup_cmd[1],
        data,
        offset,
        bit_count,
    )
    bandc, bit_count = extract_band(
        tnr_setup_cmd[4],
        do_analysis_cmd[4],
        do_analysis_cmd[3],
        tnr_setup_cmd[2],
        tnr_setup_cmd[1],
        data,
        offset,
        bit_count,
    )
    bandd, bit_count = extract_band(
        tnr_setup_cmd[3],
        do_analysis_cmd[4],
        do_analysis_cmd[3],
        tnr_setup_cmd[2],
        tnr_setup_cmd[1],
        data,
        offset,
        bit_count,
    )
    # compute the new offset for the block
    offset += bit_count // 8

    # align the address
    # I do not understand why this is a 32 bits alignment in the
    # documentation, surely I misunderstand something... but working !
    if offset % 2 != 0:
        offset += 1

    # tuple of values to add them to the numpy array
    return (
        cmd_ctr, rpw_status_cmd, temperatures, tnr_setup_cmd,
        input_setup_cmd, do_analysis_cmd, deltas,
        (banda, bandb, bandc, bandd), time
    ), offset


cpdef list create_dtype():
    """
    Create dtype of numpy for the special packet.
    """
    return [
        ("cmd_ctr", "uint8"),
        ("rpw_status", rpw_status_dtype()),
        ("temperatures", temperatures_dtype()),
        ("tnr_setup", tnr_setup_dtype()),
        ("input_setup", input_setup_dtype()),
        ("do_analysis", do_analysis_dtype()),
        ("delta_times", delta_times_dtype()),
        ("data", block_data_dtype()),
        ("acquisition_time", "uint32", 3),
    ]


cpdef list tnr_setup_dtype():
    """
    Return the TNR setup command dtype.
    """
    return [
        ("av", "uint8"), ("cr", "uint8"), ("au", "uint8"),
        ("d", "uint8"), ("c", "uint8"), ("b", "uint8"), ("a", "uint8"),
    ]


cpdef list delta_times_dtype():
    return [
        ("a", "uint32"), ("b", "uint32"), ("c", "uint32"), ("d", "uint32"),
    ]


cpdef list channel_dtype():
    return [
        ("agc", "uint16"), ("autos", "uint16", 32),
    ]


cpdef list band_dtype():
    return [
        ("ch1", channel_dtype()),
        ("ch2", channel_dtype()),
        ("cross", "uint16", (32, 2)),
    ]


cpdef list block_data_dtype():
    return [
        ("a", band_dtype()), ("b", band_dtype()),
        ("c", band_dtype()), ("d", band_dtype()),
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['roc',
 'roc.rap',
 'roc.rap.config',
 'roc.rap.tasks',
 'roc.rap.tasks.bia',
 'roc.rap.tasks.lfr',
 'roc.rap.tasks.lfr.bp',
 'roc.rap.tasks.tds',
 'roc.rap.tasks.thr',
 'roc.rap.tests',
 'roc.rap.tools']

package_data = \
{'': ['*']}

install_requires = \
['Cython>=0.29,<0.30',
 'h5py>=3.7,<4.0',
 'numpy!=1.19.5',
 'pandas>=1.1.3',
 'poppy-core',
 'poppy-pop',
 'roc-dingo>=1.0,<2.0',
 'roc-idb>=1.0,<2.0',
 'roc-rpl>=1.0,<2.0',
 'spice_manager']

setup_kwargs = {
    'name': 'roc-rap',
    'version': '1.4.2',
    'description': 'Rpw dAta Processor (RAP): a plugin used to process RPW L0, L1 and HK data',
    'long_description': 'RAP PLUGIN README\n=================\n\n[![pipeline status](https://gitlab.obspm.fr/ROC/Pipelines/Plugins/RAP/badges/develop/pipeline.svg)](https://gitlab.obspm.fr/ROC/Pipelines/Plugins/RAP/pipelines)\n\nThis directory contains the source files of the Rpw dAta Processor (RAP), a plugin of the ROC pipelines dedicated to process RPW L0 data into L1.\n\nRPL has been developed with the [POPPY framework](https://poppy-framework.readthedocs.io/en/latest/).\n\n## Quickstart\n\n### Installation with pip\n\nTo install the plugin using pip:\n\n```\npip install roc-rap --extra-index-url https://__token__:<your_personal_token>@gitlab.obspm.fr/api/v4/projects/2052/packages/pypi/simple --trusted-host gitlab.obspm.fr\n```\n\nYou will need a personnal access token to reach the package registry in the ROC Gitlab server.\n\n### Installation from the repository\n\nFirst, retrieve the `RAP` repository from the ROC gitlab server:\n\n```\ngit clone https://gitlab.obspm.fr/ROC/Pipelines/Plugins/RAP.git\n```\n\nThen, install the package (here using (poetry)[https://python-poetry.org/]):\n\n```\npoetry install --extras "poppy plugin"\n```\n\nNOTES:\n\n    - It is also possible to clone the repository using SSH\n    - To install poetry: `pip install poetry`\n\n## Usage\n\nThe roc-rap plugin is designed to be run in a POPPy-built pipeline.\nNevertheless, it is still possible to import some classes and methods in Python files.\n\n## Known issues\n\n### Installation in MacOS based on Apple silicon processors\n\nThe RAP plugin cannot be installed for the moment in MacOs relaying on Apple silicon processors.\n\n## Contacts\n\n- xavier.bonnin@obspm.fr\n- florence.henry@obspm.fr\n',
    'author': 'Xavier Bonnin',
    'author_email': 'xavier.bonnin@obspm.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.obspm.fr/ROC/Pipelines/Plugins/RAP',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



import numpy as np

from poppy.core.tools.exceptions import print_exception
from roc.rpl.packet_structure.data import Data
from roc.rpl.packet_structure.data cimport Data

from roc.idb.converters.cdf_fill import fill_records as filler

# Size in bytes of one block in PA_THR_SP_DATA parameter of TM_THR_SCIENCE_SPECTRAL_POWER packet
BLOCK_SIZE = 8

cpdef decommute(group, task):
    cdef:
        Py_ssize_t i
    """ 
    Extract source_data blocks in TM_THR_SCIENCE_SPECTRAL_POWER packets
    """

    # get the number of TM_THR_SCIENCE_SPECTRAL_POWER packets
    packet_number = group["PA_THR_SP_DATA_CNT"].shape[0]

    # get the total number of records over all TM_THR_SCIENCE_SPECTRAL_POWER packets
    record_number = int(np.sum(group["PA_THR_SP_DATA_CNT"]) / 4.0)

    # create the array containing the data
    array = np.empty(record_number, dtype=create_dtype())

    # Initialize the array with excepted CDF FILLVAL values
    filler(array)

    # loop over TM_THR_SCIENCE_SPECTRAL_POWER packets
    counter = 0
    for i in range(packet_number):
        try:
            # Get acquisition time (CUC with coarse + fine + synchro flag)
            acquisition_time = group["PA_THR_ACQUISITION_TIME"][i]

            # Number of blocks of 16 bits for the current packet
            nblock = int(group["PA_THR_SP_DATA_CNT"][i])

            # transform the data into byte array
            # store because of the garbage collector removing reference while
            # processing in Cython
            byte = group[
                "PA_THR_SP_DATA"
            ][i][:nblock].byteswap().newbyteorder().tobytes()
            data = Data(byte, len(byte))

            # Number of blocks of 64 bits in the packet
            nblock_64 = int(nblock / 4.0)

            # init offset (in bytes)
            offset = 0

            # loop over blocks
            for j in range(nblock_64):
                # now extract the data
                array[counter], offset = extract_data(
                    data,
                    offset,
                )

                counter += 1

            # check offset size
            if offset != nblock_64 * BLOCK_SIZE:
                raise Exception("Analysis has failed for TM_THR_SCIENCE_SPECTRAL_POWER")

            # Add acquisition time fine part and synchro flag for the current packet
            array[counter-nblock_64]['fine'] = group["PA_THR_ACQUISITION_TIME"][i][1]
            array[counter-nblock_64]['synchro'] = group["PA_THR_ACQUISITION_TIME"][i][2]

        except:
            # continue if something wrong with the packet
            message = print_exception()
            task.exception(message)

    return array


cdef tuple extract_data(Data data, Py_ssize_t offset):
    """
    Extract the data and insert it into the data structure of the packet.
    """
    cdef:
        unsigned long coarse
        unsigned short agc
        unsigned short auto
        unsigned char index

    # read coarse time value
    coarse = data.u32(offset, 0, 32)

    # read agc value
    agc = data.u16(offset + 4, 0, 12)

    # read auto value
    auto = data.u16(offset + 5, 4, 12)

    # read index frequency
    index = data.u8(offset + 7, 0, 8)

    # compute the new offset in bytes for the block
    offset += BLOCK_SIZE

    # tuple of values to add them to the numpy array
    # Fine and synchro not provided (FILLVAL value)
    return (
        coarse, agc, auto, index, 65535, 255,
    ), offset


cpdef list create_dtype():
    """
    Create dtype of numpy for the special packet.
    """
    return [
        ("coarse", "uint32"),
        ("agc", "uint16"),
        ("auto", "uint16"),
        ("index", "uint8"),
        ("fine", "uint16"),
        ("synchro", "uint8"),
    ]
